import PQueue from "p-queue";
import pDefer from "p-defer";
import { MemoryTree, MemoryStore } from "@canvas-js/okra-memory";
import { AbstractGossipLog } from "../AbstractGossipLog.js";
import { assert } from "../utils.js";
export class GossipLog extends AbstractGossipLog {
    static async open(init) {
        const messages = await MemoryTree.open();
        const parents = new MemoryStore();
        const ancestors = new MemoryStore();
        const gossipLog = new GossipLog(messages, parents, ancestors, init);
        if (init.replay) {
            await gossipLog.replay();
        }
        return gossipLog;
    }
    constructor(messages, parents, ancestors, init) {
        super(init);
        this.messages = messages;
        this.parents = parents;
        this.ancestors = ancestors;
        this.queue = new PQueue({ concurrency: 1 });
        this.incomingSyncPeers = new Set();
        this.outgoingSyncPeers = new Set();
    }
    async close() {
        this.log("closing");
        this.queue.clear();
        await this.queue.onIdle();
        await this.messages.store.close();
        await this.parents.close();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        const deferred = pDefer();
        this.queue.add(() => {
            return deferred.promise;
        });
        try {
            for await (const node of this.messages.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            deferred.resolve();
        }
    }
    async read(callback, options = {}) {
        const targetPeerId = options.target ?? null;
        if (targetPeerId !== null) {
            if (this.outgoingSyncPeers.has(targetPeerId.toString())) {
                throw new Error(`deadlock with peer ${targetPeerId}`);
            }
        }
        const result = await this.queue.add(async () => {
            if (targetPeerId !== null) {
                this.incomingSyncPeers.add(targetPeerId.toString());
            }
            try {
                return await callback({ messages: this.messages, parents: this.parents, ancestors: this.ancestors });
            }
            catch (err) {
                this.log.error("error in transaction: %O", err);
            }
            finally {
                if (targetPeerId !== null) {
                    this.incomingSyncPeers.delete(targetPeerId.toString());
                }
            }
        });
        return result;
    }
    async write(callback, options = {}) {
        const sourcePeerId = options.source ?? null;
        if (sourcePeerId !== null) {
            if (this.incomingSyncPeers.has(sourcePeerId.toString())) {
                throw new Error(`deadlock with peer ${sourcePeerId}`);
            }
        }
        const result = await this.queue.add(async () => {
            if (sourcePeerId !== null) {
                this.outgoingSyncPeers.add(sourcePeerId.toString());
            }
            try {
                return await callback({ messages: this.messages, parents: this.parents, ancestors: this.ancestors });
            }
            catch (err) {
                this.log.error("error in transaction: %O", err);
                throw err;
            }
            finally {
                if (sourcePeerId !== null) {
                    this.outgoingSyncPeers.delete(sourcePeerId.toString());
                }
            }
        });
        return result;
    }
}
