import { varint } from "multiformats";
import { base32hex } from "multiformats/bases/base32";
import { sha256 } from "@noble/hashes/sha256";
import * as cbor from "@ipld/dag-cbor";
import { fromDSL } from "@ipld/schema/from-dsl.js";
import { create } from "@ipld/schema/typed.js";
import { assert } from "./utils.js";
const schema = fromDSL(`
type SignedMessage struct {
	signature nullable Signature
	message Message
} representation tuple

type Signature struct {
	type String
	publicKey Bytes
	signature Bytes
	cid &Message
} representation tuple

type Message struct {
	parents nullable [Bytes]
	payload any
} representation tuple
`);
const { toTyped: toSignedMessage, toRepresentation: fromSignedMessage } = create(schema, "SignedMessage");
export function decodeSignedMessage(topic, value) {
    const { signature, message } = toSignedMessage(cbor.decode(value));
    const clock = getClock(message.parents);
    const key = getKey(clock, sha256(value));
    const parents = message.parents?.map(decodeId) ?? [];
    assert(parents.every((id, i) => i === 0 || parents[i - 1] < id), "unsorted parents array");
    return [decodeId(key), signature, { topic, clock, parents, payload: message.payload }];
}
export function encodeSignedMessage(signature, { clock, parents, payload }) {
    parents.sort();
    const signedMessage = {
        signature,
        message: {
            parents: clock === 0 ? null : parents.map(encodeId),
            payload,
        },
    };
    const value = cbor.encode(fromSignedMessage(signedMessage));
    const key = getKey(clock, sha256(value));
    return [key, value];
}
export function getClock(parents) {
    if (parents === null) {
        return 0;
    }
    let max = 0;
    for (const key of parents) {
        assert(key.byteLength === KEY_LENGTH, "expected key.byteLength === KEY_LENGTH");
        const [clock] = varint.decode(key);
        if (clock > max) {
            max = clock;
        }
    }
    return max + 1;
}
export const KEY_LENGTH = 20;
function getKey(clock, hash) {
    const encodingLength = varint.encodingLength(clock);
    const key = new Uint8Array(KEY_LENGTH);
    varint.encodeTo(clock, key, 0);
    key.set(hash.subarray(0, key.byteLength - encodingLength), encodingLength);
    return key;
}
export const encodeId = (id) => base32hex.baseDecode(id);
export const decodeId = (key) => base32hex.baseEncode(key);
